"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const child_process = require("child_process");
const events_1 = require("events");
const path = require("path");
const util_1 = require("util");
const checkSanity_1 = require("../../shared/checkSanity");
const Util_1 = require("../../shared/Util");
const ManagedChildProcess_1 = require("../ManagedChildProcess");
const BackgroundServicesFile_1 = require("./BackgroundServicesFile");
const execFile = util_1.promisify(child_process.execFile);
class BackgroundServices extends events_1.EventEmitter {
    constructor() {
        super(...arguments);
        /** If the .start() method has been called and is done */
        this.isStartDone = false;
    }
    /** Start all required background process for this platform */
    async start(config) {
        const logOutput = (entry) => { this.log(entry); };
        // Keep some configs
        this.flashpointPath = config.flashpointPath;
        this.useFiddler = config.useFiddler;
        // Abort if Flashpoint path is not valid
        const valid = await checkSanity_1.isFlashpointValidCheck(config.flashpointPath);
        if (!valid) {
            this.logContent('The flashpoint root folder is invalid.');
            this.startDone();
            return;
        }
        // Load background service info from file
        let serviceInfo;
        try {
            const jsonFolder = path.posix.join(config.flashpointPath, config.jsonFolderPath);
            serviceInfo = await BackgroundServicesFile_1.BackgroundServicesFile.readFile(jsonFolder, error => this.logContent(error));
        }
        catch (error) {
            this.logContent(`An error occurred while loading the background services file:\n  ${error.toString()}`);
            this.startDone();
            return;
        }
        //console.log('Services:', serviceInfo);
        this.serviceInfo = serviceInfo;
        // Run start commands
        const start = serviceInfo.start;
        for (let i = 0; i < start.length; i++) {
            await this.execProcess(start[i]);
        }
        // Start router
        if (config.startRouter) {
            if (!serviceInfo.server) {
                throw new Error('Server process information not found.');
            }
            this.server = createManagedChildProcess('Router', serviceInfo.server);
            this.server.on('output', logOutput);
            this.spawnProc(this.server);
        }
        // Start redirector
        // (Linux doesn't have to do this)
        if (config.startRedirector && process.platform !== 'linux') {
            const redirectorInfo = config.useFiddler ? serviceInfo.fiddler : serviceInfo.redirector;
            if (!redirectorInfo) {
                throw new Error(`Redirector process information not found. (Type: ${config.useFiddler ? 'Fiddler' : 'Redirector'})`);
            }
            this.redirector = createManagedChildProcess('Redirector', redirectorInfo, config.useFiddler);
            this.redirector.on('output', logOutput);
            this.spawnProc(this.redirector);
        }
        this.startDone();
        // -- Functions --
        // Wrapper for ManagedChildProcess's constructor
        function createManagedChildProcess(name, info, detached) {
            return new ManagedChildProcess_1.ManagedChildProcess(name, info.filename, info.arguments, path.join(config.flashpointPath, info.path), !!detached);
        }
    }
    startDone() {
        // Update start done flag and emit event
        this.isStartDone = true;
        this.emit('start-done');
    }
    /** Stop all currently active background processes */
    async stop() {
        if (!this.startDone) {
            throw new Error('You must not stop the background services before they are done starting.');
        }
        if (!this.serviceInfo) {
            return;
        }
        // Kill processes
        if (this.server) {
            if (this.serviceInfo.server && this.serviceInfo.server.kill) {
                this.server.kill();
            }
        }
        if (this.redirector) {
            const doKill = !!(this.useFiddler ?
                this.serviceInfo.fiddler && this.serviceInfo.fiddler.kill :
                this.serviceInfo.redirector && this.serviceInfo.redirector.kill);
            if (doKill) {
                this.redirector.kill();
            }
        }
        // Run stop commands
        const stop = this.serviceInfo.stop;
        for (let i = 0; i < stop.length; i++) {
            await this.execProcess(stop[i], true);
        }
    }
    /** Wait until this is done starting (doesn't wait if already done starting) */
    async waitUntilDoneStarting() {
        // Check if already done starting
        if (this.isStartDone) {
            return;
        }
        // Wait for the start done event
        await new Promise((resolve) => {
            this.once('start-done', () => { resolve(); });
        });
    }
    /** Execute a process and wait for it to finish */
    async execProcess(proc, sync) {
        if (this.flashpointPath === undefined) {
            throw new Error('BackgroundServices#flashpointPath must not be undefined when executing a process');
        }
        const cwd = path.join(this.flashpointPath, proc.path);
        this.logContent(`Executing "${proc.filename}" ${Util_1.stringifyArray(proc.arguments)} in "${proc.path}"`);
        try {
            if (sync) {
                child_process.execFileSync(proc.filename, proc.arguments, { cwd: cwd });
            }
            else {
                await execFile(proc.filename, proc.arguments, { cwd: cwd });
            }
        }
        catch (error) {
            this.logContent(`An unexpected error occurred while executing a command:\n  "${error}"`);
        }
    }
    log(entry) {
        this.emit('output', entry);
    }
    logContent(content) {
        this.emit('output', {
            source: 'Background Services',
            content,
        });
    }
    /** Try to spawn a ManagedChildProcess, and log error if it fails */
    spawnProc(proc) {
        try {
            proc.spawn();
        }
        catch (error) {
            this.logContent(`An unexpected error occurred while trying to run the background process "${proc.name}".` +
                `  ${error.toString()}`);
        }
    }
}
exports.default = BackgroundServices;
//# sourceMappingURL=BackgroundServices.js.map