"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const electron_1 = require("electron");
const events_1 = require("events");
const AppPreferencesApi_1 = require("../../shared/preferences/AppPreferencesApi");
const util_1 = require("../../shared/preferences/util");
const AppPreferencesFile_1 = require("./AppPreferencesFile");
/**
 * Manager of the Preferences Data.
 * "Back end" of the Preferences API, this lives in the "Main" process.
 * This is the bridge between "AppPreferencesApi" and the Preferences file.
 */
class AppPreferencesMain extends events_1.EventEmitter {
    get data() {
        if (!this._data) {
            throw new Error('You must not try to access the preferences data before it is loaded!');
        }
        return this._data;
    }
    constructor() {
        super();
        // Add IPC event listeners
        electron_1.ipcMain
            .on(AppPreferencesApi_1.AppPreferencesApi.ipcSend, this.onSendData.bind(this))
            .on(AppPreferencesApi_1.AppPreferencesApi.ipcRequestSync, this.onRequestDataSync.bind(this));
    }
    /** Load the data from the file */
    async load() {
        this._data = await AppPreferencesFile_1.AppPreferencesFile.readOrCreate(this.log.bind(this));
        console.log('Preferences:', this._data);
    }
    /** Called when data is sent to this from the renderers preferences api */
    onSendData(event, data) {
        if (!this._data) {
            throw new Error('The data must first be loaded before new can be received.');
        }
        if (!data) {
            throw new Error('You must send a data object, but no data was received.');
        }
        // Update the data
        this._data = util_1.overwritePreferenceData(this._data, data, this.log.bind(this));
        // Save the data
        AppPreferencesFile_1.AppPreferencesFile.saveFile(this._data);
        // Send response (so the renderer knows that the data was received)
        event.sender.send(AppPreferencesApi_1.AppPreferencesApi.ipcSendResponse);
    }
    /** Called when data is requested from the renderers preferences api */
    onRequestDataSync(event) {
        event.returnValue = this._data;
    }
    log(content) {
        this.emit('log', { source: 'Preferences', content });
    }
}
exports.AppPreferencesMain = AppPreferencesMain;
//# sourceMappingURL=AppPreferencesMain.js.map