"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const path = require("path");
/**
 * Read and parse a JSON file asynchronously.
 * Wrapper around "fs.readFile()" plus "JSON.parse()".
 * @param path Path of the JSON file
 * @param options Options for reading the file
 */
function readJsonFile(path, options) {
    return new Promise((resolve, reject) => {
        fs.readFile(path, options, (error, data) => {
            // Check if reading file failed
            if (error) {
                return reject(error);
            }
            // Try to parse json (and callback error if it fails)
            const jsonOrError = tryParseJSON(data);
            if (jsonOrError instanceof Error) {
                return reject(jsonOrError);
            }
            // Success!
            return resolve(jsonOrError);
        });
    });
}
exports.readJsonFile = readJsonFile;
;
/**
 * Remove the file extension of a filename
 * (Remove everything after the last dot, including the dot)
 */
function removeFileExtension(filename) {
    const lastDotIndex = filename.lastIndexOf('.');
    if (lastDotIndex === -1) {
        return filename;
    }
    return filename.substr(0, lastDotIndex);
}
exports.removeFileExtension = removeFileExtension;
/**
 * Get the filename of a path or url
 * (get everything after the last slash symbol)
 * @param filePath Path to get filename from
 */
function getFilename(filePath) {
    return filePath.substr(Math.max(filePath.lastIndexOf('/'), filePath.lastIndexOf('\\')) + 1);
}
exports.getFilename = getFilename;
/**
 * Pad a the end of a string with spaces until the string is of a specified length
 * @param str String to pad
 * @param length Target length of string (max number of spaces to add)
 * @returns String padded with spaces
 *          (or the original string if it's length is equal or longer than the specified length)
 */
function padEnd(str, length) {
    str = str + ''; // (Coerce to string)
    return str + ' '.repeat(Math.max(0, length - str.length));
}
exports.padEnd = padEnd;
/**
 * Pad a the start of a string with spaces until the string is of a specified length
 * @param str String to pad
 * @param length Target length of string (max number of spaces to add)
 * @returns String padded with spaces
 *          (or the original string if it's length is equal or longer than the specified length)
 */
function padStart(str, length) {
    str = str + ''; // (Coerce to string)
    return ' '.repeat(Math.max(0, length - str.length)) + str;
}
exports.padStart = padStart;
/**
 * Write an array to a string in a pretty and readable way
 * Ex. [0,'test',null] => "[ 0, 'test', null ]"
 * @param array Array to "stringify"
 * @returns Readable text representation of the array
 */
function stringifyArray(array, opts) {
    const trimStrings = opts && opts.trimStrings || false;
    // Build string
    let str = '[ ';
    for (let i = 0; i < array.length; i++) {
        let element = array[i];
        if (isString(element)) {
            str += `"${trimStrings ? trim(element) : element}"`;
        }
        else {
            str += element;
        }
        if (i !== array.length - 1) {
            str += ', ';
        }
    }
    str += ' ]';
    return str;
}
exports.stringifyArray = stringifyArray;
/** Remove all spaces and new line characters at the start and end of the string. */
function trim(str) {
    let first = 0;
    let last = str.length;
    // Find the first non-space non-new-line character
    for (let i = 0; i < str.length; i++) {
        if (!isSpaceOrNewLine(str[i])) {
            first = i;
            break;
        }
    }
    // Find the last non-space non-new-line character
    for (let i = str.length - 1; i >= first; i--) {
        if (!isSpaceOrNewLine(str[i])) {
            last = i;
            break;
        }
    }
    // Get the character between the first and last (including them)
    return str.substring(first, last);
}
function isSpaceOrNewLine(char) {
    switch (char) {
        case ' ':
        case '\n': return true;
        default: return false;
    }
}
/**
 * Get the ISO formatted time stamp from a date object.
 * ("yyyy-MM-ddThh:mm:ss.fff+00:00")
 */
function formatDate(d) {
    return `${pad(d.getFullYear(), 4)}-${pad(d.getMonth() + 1, 2)}-${pad(d.getDate(), 2)}` +
        `T${pad(d.getHours(), 2)}:${pad(d.getMinutes(), 2)}:${pad(d.getSeconds(), 2)}` +
        `.${pad(d.getMilliseconds(), 3)}${timezone(d.getTimezoneOffset())}`;
}
exports.formatDate = formatDate;
// (Pads the beginning of a string with "0"s until it reaches a specified length)
function pad(str, len) {
    return '0'.repeat(Math.max(0, len - (str + '').length)) + str;
}
// (Converts a timestamp (in minutes, not milliseconds) to the timezone part of the ISO date string ("(+/-)hh:mm"))
function timezone(time) {
    const t = Math.abs(time);
    return `${(time < 0) ? '+' : '-'}${pad(Math.floor(t / 60), 2)}:${pad(t % 60, 2)}`;
}
/**
 * Stringify anything to a json string ready to be saved to a file
 * @param data Data to be stringified
 * @returns JSON string of given data
 */
function stringifyJsonDataFile(data) {
    return JSON.stringify(data, null, 2).replace(/\n/g, '\r\n');
}
exports.stringifyJsonDataFile = stringifyJsonDataFile;
/**
 * Check if all properties of both arguments have strictly equals values,
 * and if both objects have identical properties (same number of props with the same names)
 * @param first
 * @param second
 */
function shallowStrictEquals(first, second) {
    for (let key in first) {
        if (!(key in second) || first[key] !== second[key]) {
            return false;
        }
    }
    for (let key in second) {
        if (!(key in first) || first[key] !== second[key]) {
            return false;
        }
    }
    return true;
}
exports.shallowStrictEquals = shallowStrictEquals;
/**
 * Recursively copy values from data to target (for every property of the same name)
 * @param target Target object to copy data to
 * @param source Source object to copy data from
 * @returns Target object
 */
function recursiveReplace(target, source) {
    // Skip if source is missing
    if (!source) {
        return target;
    }
    // Go through all properties of target
    for (let key in source) {
        // Check if data has a property of the same name
        if (key in target) {
            const val = source[key];
            // If the value is an object
            if (val !== null && typeof val === 'object') {
                // Go one object deeper and continue copying
                recursiveReplace(target[key], val);
            }
            else {
                // Copy the value
                target[key] = val;
            }
        }
    }
    return target;
}
exports.recursiveReplace = recursiveReplace;
/**
 * Recursively copy and object and its "sub-objects"
 * (WARNING: This will overflow the stack if it tries to copy circular references)
 * @param source Object to copy from
 * @returns New copy of source
 */
function deepCopy(source) {
    const copy = Array.isArray(source) ? [] : {};
    for (let key in source) {
        let val = source[key];
        if (val !== null && typeof val === 'object') {
            val = deepCopy(val);
        }
        copy[key] = val;
    }
    return copy;
}
exports.deepCopy = deepCopy;
/** Try parsing a JSON string into an object and return that object, or an error if one occurred */
function tryParseJSON(jsonString) {
    let ret;
    try {
        ret = JSON.parse(jsonString);
    }
    catch (error) {
        ret = error;
    }
    return ret;
}
exports.tryParseJSON = tryParseJSON;
/**
 * Recursively go down a folder and call back for each file encountered
 * @param options Various options (a shallow copy of this is accessible from the callback)
 * @returns A promise that resolves when either all files have been called back for or recursion is aborted
 */
async function recursiveDirectory(options) {
    const shared = {
        options: Object.assign({}, options),
        abort: false,
    };
    return innerRecursiveDirectory(shared, '');
}
exports.recursiveDirectory = recursiveDirectory;
async function innerRecursiveDirectory(shared, dirPath) {
    return new Promise(async function (resolve, reject) {
        // Full path to the current folder
        const fullDirPath = path.join(shared.options.directoryPath, dirPath);
        // Get the names of all files and sub-folders
        fs.readdir(fullDirPath, function (err, files) {
            if (shared.abort) {
                return resolve();
            } // (Abort exit point)
            if (err) {
                reject(err);
            }
            else {
                // Resolve if folder is empty
                if (files.length === 0) {
                    return resolve();
                }
                // Get the stats of each folder/file to verify if they are a folder or file
                // (And wait for every single one to complete before resolving the promise)
                let filesOrFoldersLeft = files.length;
                for (let i = files.length - 1; i >= 0; i--) {
                    const filename = files[i];
                    fs.stat(path.join(fullDirPath, filename), async function (err, stats) {
                        if (shared.abort) {
                            return resolve();
                        } // (Abort exit point)
                        if (err) {
                            reject(err);
                        }
                        else {
                            if (stats.isFile()) {
                                const p = shared.options.fileCallback({
                                    shared: shared,
                                    filename: filename,
                                    relativePath: dirPath,
                                });
                                if (p) {
                                    await p;
                                }
                            }
                            else {
                                await innerRecursiveDirectory(shared, path.join(dirPath, filename)).catch(reject);
                            }
                        }
                        filesOrFoldersLeft -= 1;
                        if (filesOrFoldersLeft === 0) {
                            resolve();
                        }
                    });
                }
            }
        });
    });
}
function isString(obj) {
    return typeof obj === 'string' || obj instanceof String;
}
//# sourceMappingURL=Util.js.map