"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const electron_1 = require("electron");
const events_1 = require("events");
const Util_1 = require("../Util");
const util_1 = require("./util");
/**
 * Bridge between the Renderer and "AppPreferencesMain" (which then accesses the Preferences file).
 * "Front end" of the Preferences API, this lives in the "Renderer" process.
 * This is initially handed the preferences loaded from the file (or the defaults),
 * it then checks if the renderer changes the preferences and sends them to the main at an interval.
 */
class AppPreferencesApi extends events_1.EventEmitter {
    constructor() {
        super(...arguments);
        /** If this is initialized */
        this._isInit = false;
        /** If the data has changed since the last time the data was sent */
        this._dataChanged = false;
        /** If data has been sent, and no response has yet been received */
        this._isSending = false;
    }
    /**
     * Initialize (this should be called after construction, and before accessing the data object)
     */
    async initialize() {
        return new Promise(async () => {
            if (this._isInit) {
                throw new Error('You can only initialize this once');
            }
            // Fetch initial preferences data from main
            const data = await this.fetch();
            // Keep data
            this._dataCache = Util_1.deepCopy(data);
            // Create proxy for data object
            this._dataProxy = new Proxy(this._dataCache, {
                // Whenever the value of a data property is set
                set: (target, p, value, receiver) => {
                    // Check if the property's value was changed
                    if (target[p] !== value) {
                        this._dataChanged = true;
                    }
                    // Set property's value as normal
                    return Reflect.set(target, p, value, receiver);
                },
            });
            // Start send loop
            setInterval(() => {
                if (this._dataChanged) {
                    this._dataChanged = false;
                    this.send();
                }
            }, AppPreferencesApi.sendDataInterval);
            // Send data when window is closing
            window.addEventListener('unload', () => {
                this.send();
            });
            // Done
            this._isInit = true; // Update Flag
            this.emit('init'); // Emit event
        });
    }
    /** Wait until this is initialized (doesn't wait if already initialized) */
    async waitUtilInitialized() {
        // Check if already initialized
        if (this._isInit) {
            return;
        }
        // Wait for the init event
        await new Promise((resolve) => {
            this.once('init', () => { resolve(); });
        });
    }
    /**
     * Send the current preference data to the main process
     * @returns If the send was successful
     */
    async send() {
        if (this._isSending) {
            return false;
        }
        this._isSending = true;
        // Send data and wait for response
        return new Promise((resolve, reject) => {
            // @TODO Add a timeout check (reject if it hasn't responded for something like 15 sec)
            electron_1.ipcRenderer.once(AppPreferencesApi.ipcSendResponse, () => {
                this._isSending = false; // Update flag
                resolve(true);
            });
            electron_1.ipcRenderer.send(AppPreferencesApi.ipcSend, this._dataCache);
        });
    }
    /** Fetch the preference data from the main process */
    async fetch() {
        // Send data and wait for response
        return new Promise((resolve, reject) => {
            const data = electron_1.ipcRenderer.sendSync(AppPreferencesApi.ipcRequestSync);
            if (data) {
                resolve(data);
            }
            else {
                reject(new Error('No data received from preference data fetch request'));
            }
        });
    }
    /** Get the currently cached data (wrapped in a proxy) */
    getData() {
        if (!this._dataProxy) {
            throw new Error('You must not call AppPreferencesApi.getData before it has loaded');
        }
        return this._dataProxy;
    }
    /** Set the data */
    setData(data) {
        if (!this._dataProxy) {
            throw new Error('You must not call AppPreferencesApi.setData before it has loaded');
        }
        util_1.overwritePreferenceData(this._dataProxy, data);
    }
}
/** How often the data should be sent to the main (in milliseconds) */
AppPreferencesApi.sendDataInterval = 0.5 * 1000;
/** Send Preferences Data (renderer -> main) (IPC Event Name) */
AppPreferencesApi.ipcSend = 'app-preferences-api-send';
/** Response to sent Preferences Data (main -> renderer) (IPC Event Name) */
AppPreferencesApi.ipcSendResponse = 'app-preferences-api-send-response';
/** Request Preferences Data (renderer -> main) (IPC Event Name) */
AppPreferencesApi.ipcRequestSync = 'app-preferences-api-request-sync';
exports.AppPreferencesApi = AppPreferencesApi;
//# sourceMappingURL=AppPreferencesApi.js.map